from git.repo import Repo
from git.exc import InvalidGitRepositoryError

import os

CURRDIR = os.getcwd()
SHADOW_BRANCH_PREFIX = "flor."


def get_repo_dir():
    try:
        repo = Repo(CURRDIR, search_parent_directories=True)
        return repo.working_dir
    except InvalidGitRepositoryError:
        print("Not a valid Git repository")
    except Exception as e:
        print(f"An error occurred while getting the repository directory: {e}")


def git_commit(message="FLOR::Auto-commit"):
    try:
        # Get the current working directory and initialize a Repo object
        repo = Repo(CURRDIR, search_parent_directories=True)

        # Check if there are any uncommitted changes
        if repo.is_dirty(untracked_files=True):
            # Add all untracked files and changes to tracked files
            repo.git.add(A=True)

            # Commit the changes
            repo.git.commit(m=message)
            print("Changes committed successfully.")
        else:
            print("No changes to commit.")
    except InvalidGitRepositoryError:
        print("Not a valid Git repository")
    except Exception as e:
        print(f"An error occurred while committing: {e}")


def current_branch():
    try:
        repo = Repo(CURRDIR, search_parent_directories=True)
        return repo.active_branch
    except InvalidGitRepositoryError:
        return None
    except TypeError:
        return None


def to_shadow():
    try:
        repo = Repo(CURRDIR, search_parent_directories=True)
        branch = repo.active_branch.name
        if branch.startswith(SHADOW_BRANCH_PREFIX):
            # Branch already has the 'flor.' prefix, continuing...
            return
        else:
            base_shadow_name = "flor.shadow"
            new_branch_name = base_shadow_name
            suffix = 1

            # Check if the branch name exists and increment the suffix until a unique name is found
            while any(b.name == new_branch_name for b in repo.branches):
                new_branch_name = f"{base_shadow_name}{suffix}"
                suffix += 1

            try:
                # Try to create a new branch with the unique name
                repo.git.checkout("-b", new_branch_name)
                print(f"Created and switched to new branch: {new_branch_name}")
            except Exception as e:
                # Likely branch already exists due to race condition
                # repo.git.checkout(new_branch_name)
                branch = repo.active_branch.name
                print(
                    f"Branch '{new_branch_name}' already exists. Switched to branch: {branch}"
                )
    except InvalidGitRepositoryError:
        print("Not a valid Git repository")
    except Exception as e:
        print(f"An error occurred while processing the branch: {e}")


def get_latest_autocommit():
    try:
        repo = Repo(CURRDIR, search_parent_directories=True)
        for v in repo.iter_commits():
            if str(v.message).count("FLOR::") == 1:
                _, _, ts = v.message.strip().split("::")  # type: ignore
                yield (
                    str(ts),
                    v.hexsha,
                    v.authored_datetime.isoformat(timespec="seconds")[0 : len(ts)],
                )
    except InvalidGitRepositoryError:
        print("Not a valid Git repository")
    except Exception as e:
        print(f"An error occurred while processing the branch: {e}")


def checkout(commit_hash):
    repo = Repo(CURRDIR, search_parent_directories=True)
    # Checkout to the desired commit
    print("Checking out ", commit_hash)
    repo.git.checkout(commit_hash)


def get_head():
    repo = Repo(CURRDIR, search_parent_directories=True)
    current_head = repo.head.commit
    return current_head


def reset_hard():
    repo = Repo(CURRDIR, search_parent_directories=True)
    repo.git.reset("--hard")



#!/usr/bin/env python
# type: ignore
from itertools import zip_longest

from . import GumTree, Mapping
from .tree import Node, adapter


N = Node
gt = GumTree(adapter)


def test():
    t1, t2 = example()

    result = gt.topdown(t1, t2)

    expected = Mapping(adapter)
    expected.put_tree(t1[0][2][1], t2[0][2][1])
    expected.put_tree(t1[0][2][3], t2[0][2][3])
    expected.put_tree(t1[0][2][4][0][0], t2[0][2][4][0][0])
    expected.put_tree(t1[0][2][4][0][1], t2[0][2][4][0][2][1])
    assert match(result, expected), "topdown"

    result = Mapping(adapter, expected)
    gt.bottomup(t1, t2, result)

    expected.put(t1[0][2][4][0], t2[0][2][4][0])
    expected.put(t1[0][2][4], t2[0][2][4])
    expected.put(t1[0][2], t2[0][2])
    expected.put(t1[0][2][0], t2[0][2][0])
    expected.put(t1[0][2][2], t2[0][2][2])
    expected.put(t1[0], t2[0])
    expected.put(t1[0][0], t2[0][0])
    expected.put(t1[0][1], t2[0][1])
    expected.put(t1, t2)
    assert match(result, expected), "bottomup"

    assert match(gt.mapping(t1, t2), expected)
    print("Passed Example!")


def example():
    source = N(
        "CompilationUnit",
        "",
        [
            N(
                "TypeDeclaration",
                "",
                [
                    N("Modifier", "public"),
                    N("SimpleName", "Test"),
                    N(
                        "MethodDeclaration",
                        "",
                        [
                            N("Modifier", "private"),
                            N(
                                "SimpleType",
                                "String",
                                [
                                    N("SimpleName", "String"),
                                ],
                            ),
                            N("SimpleName", "foo"),
                            N(
                                "SingleVariableDeclaration",
                                "",
                                [
                                    N("PrimitiveType", "int"),
                                    N("SimpleName", "i"),
                                ],
                            ),
                            N(
                                "Block",
                                "",
                                [
                                    N(
                                        "IfStatement",
                                        "",
                                        [
                                            N(
                                                "InfixExpression",
                                                "==",
                                                [
                                                    N("SimpleName", "i"),
                                                    N("NumberLiteral", "0"),
                                                ],
                                            ),
                                            N(
                                                "ReturnStatement",
                                                "",
                                                [
                                                    N("StringLiteral", "Foo!"),
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            )
        ],
    )
    destination = N(
        "CompilationUnit",
        "",
        [
            N(
                "TypeDeclaration",
                "",
                [
                    N("Modifier", "public"),
                    N("SimpleName", "Test"),
                    N(
                        "MethodDeclaration",
                        "",
                        [
                            N("Modifier", "private"),
                            N(
                                "SimpleType",
                                "String",
                                [
                                    N("SimpleName", "String"),
                                ],
                            ),
                            N("SimpleName", "foo"),
                            N(
                                "SingleVariableDeclaration",
                                "",
                                [
                                    N("PrimitiveType", "int"),
                                    N("SimpleName", "i"),
                                ],
                            ),
                            N(
                                "Block",
                                "",
                                [
                                    N(
                                        "IfStatement",
                                        "",
                                        [
                                            N(
                                                "InfixExpression",
                                                "==",
                                                [
                                                    N("SimpleName", "i"),
                                                    N("NumberLiteral", "0"),
                                                ],
                                            ),
                                            N(
                                                "ReturnStatement",
                                                "",
                                                [
                                                    N("StringLiteral", "Bar"),
                                                ],
                                            ),
                                            N(
                                                "IfStatement",
                                                "",
                                                [
                                                    N(
                                                        "InfixExpression",
                                                        "==",
                                                        [
                                                            N("SimpleName", "i"),
                                                            N(
                                                                "PrefixExpression",
                                                                "-",
                                                                [
                                                                    N(
                                                                        "NumberLiteral",
                                                                        "1",
                                                                    ),
                                                                ],
                                                            ),
                                                        ],
                                                    ),
                                                    N(
                                                        "ReturnStatement",
                                                        "",
                                                        [
                                                            N("StringLiteral", "Foo!"),
                                                        ],
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            )
        ],
    )
    return source, destination


def match(left: Mapping, right: Mapping):
    return all(
        id(el) == id(rl) and id(er) == id(rr)
        for (el, er), (rl, rr) in zip_longest(left.items(), right.items())
    )


if __name__ == "__main__":
    test()



import os
from copy import deepcopy
from pathlib import Path
from .constants import *
from .clock import Clock
from . import orm
from . import cli
from . import utils
from . import versions
from . import obj_store
from . import database

from typing import Any, Iterable, Iterator, TypeVar, Optional, Union
from contextlib import contextmanager

from tqdm import tqdm
import atexit
import sqlite3

T = TypeVar("T")

output_buffer = []

layers = {}
context = None

checkpoints = []
checkpointing_clock = Clock()

skip_cleanup = True


def log(name, value):
    if skip_cleanup:
        _deferred_init()

    serializable_value = value if utils.is_jsonable(value) else str(value)
    tqdm.write(utils.to_string(layers, name, serializable_value))

    if cli.in_replay_mode() and name not in cli.flags.queryparameters["VARS"]:
        # Check that name is in logging statement propagation list
        return value

    output_buffer.append(
        orm.Log(
            PROJID,
            Clock.get_datetime(),
            SCRIPTNAME,
            context if context is None else deepcopy(context),
            name,
            serializable_value,
            1,
        )
    )

    return value


def arg(name: str, default: Optional[Any] = None) -> Any:
    if cli.in_replay_mode():
        # GIT
        assert name in cli.flags.hyperparameters
        historical_v = cli.flags.hyperparameters[name]
        log(name, historical_v)
        return historical_v
    elif name in cli.flags.hyperparameters:
        # CLI
        v = cli.flags.hyperparameters[name]
        if default is not None:
            v = utils.duck_cast(v, default)
            log(name, v)
            return v
        log(name, v)
        return v
    elif default is not None:
        # default
        log(name, default)
        return default
    else:
        raise


@contextmanager
def checkpointing(**kwargs):
    try:
        # Add prefix time delta to output_buffer
        output_buffer.append(
            orm.Log(
                PROJID,
                Clock.get_datetime(),
                SCRIPTNAME,
                context if context is None else deepcopy(context),
                "delta::prefix",
                checkpointing_clock.get_delta(),
                3,
            )
        )
        # set up the context
        checkpoints.extend(list(kwargs.items()))
        # TODO: while there are func skipblocks to load
        yield
    except Exception as e:
        # Here you can handle the exception if needed, or log it, etc.
        print(f"An error occurred: {e}")
        # Optionally re-raise the exception if you want the error to propagate
        raise
    finally:
        # This code runs whether an exception occurred or not
        checkpoints.clear()
        checkpointing_clock.set_start_time()


@contextmanager
def iteration(name: str, idx: Optional[int], value: Optional[str]):
    global context
    clock = Clock()
    clock.set_start_time()
    layers[name] = (
        int(idx) if idx is not None else None,
        str(value) if value is not None else None,
    )
    parent_context = context
    if cli.in_replay_mode():
        # TODO: load the end-state checkpoint
        load_ckpt()
        raise
    else:
        context = orm.Loop(
            orm.generate_64bit_id(),
            deepcopy(parent_context) if parent_context is not None else None,
            name,
            layers[name][0],
            layers[name][1],
        )
        yield
        ckpt()
    context = parent_context
    output_buffer.append(
        orm.Log(
            PROJID,
            Clock.get_datetime(),
            SCRIPTNAME,
            deepcopy(context) if context is not None else None,
            "delta::iteration",
            clock.get_delta(),
            3,
        )
    )
    del layers[name]


def loop(name: str, iterator: Iterable[T]) -> Iterator[T]:
    global context
    clock = Clock()
    clock.set_start_time()
    pos = len(layers)
    layers[name] = (0, None)
    parent_context = context
    for each in tqdm(
        (
            enumerate(slice(name, iterator))
            if not cli.in_replay_mode()
            else slice(name, iterator)
        ),
        position=pos,
        leave=(True if pos == 0 else False),
    ):
        layers[name] = (
            int(each[0]) + 1,
            str(each[1]) if utils.is_jsonable(each[1]) else None,
        )
        context = orm.Loop(
            orm.generate_64bit_id(),
            deepcopy(parent_context) if parent_context is not None else None,
            name,
            layers[name][0],
            layers[name][1],
        )
        if pos == 0 and cli.in_replay_mode():
            load_ckpt()
        yield each[1]  # type: ignore
        if pos == 0 and not cli.in_replay_mode():
            ckpt()
    context = parent_context
    output_buffer.append(
        orm.Log(
            PROJID,
            Clock.get_datetime(),
            SCRIPTNAME,
            deepcopy(context) if context is not None else None,
            "delta::loop",
            clock.get_delta(),
            3,
        )
    )
    del layers[name]


def commit():
    global skip_cleanup
    # Add suffix time delta to output_buffer
    output_buffer.append(
        orm.Log(
            PROJID,
            Clock.get_datetime(),
            SCRIPTNAME,
            None if context is None else deepcopy(context),
            "delta::suffix",
            checkpointing_clock.get_delta(),
            3,
        )
    )
    # Add logging statements on REPLAY
    conn, cursor = database.conn_and_cursor()
    if not cli.in_replay_mode():
        # RECORD
        branch = versions.current_branch()
        if branch is not None:
            orm.to_json(output_buffer)
            database.unpack(output_buffer, cursor)
            versions.git_commit(f"FLOR::Auto-commit::{Clock.get_datetime()}")
    else:
        database.unpack(output_buffer, cursor)
    conn.commit()
    cursor = conn.cursor()
    database.deduplicate_table(cursor, "loops")
    conn.commit()
    conn.close()
    output_buffer.clear()
    Clock.set_new_datetime()
    checkpointing_clock.s_time = None
    skip_cleanup = True


@atexit.register
def cleanup():
    if skip_cleanup:
        return
    commit()


def _deferred_init():
    global skip_cleanup
    if skip_cleanup:
        skip_cleanup = False
        if not cli.in_replay_mode():
            assert (
                versions.current_branch() is not None
            ), "Running from a detached HEAD?"
            versions.to_shadow()


def ckpt():
    for name, obj in checkpoints:
        obj_store.serialize(layers, name, obj)


def load_ckpt():
    for name, obj in checkpoints:
        obj_store.deserialize(layers, name, obj)


def slice(name, iterator):
    if not cli.in_replay_mode():
        return iterator
    original = list(iterator)

    qop = (
        (cli.flags.queryparameters).get(name, 0)
        if cli.flags.queryparameters is not None
        else 0
    )
    if qop == 1:
        return enumerate(iterator)

    new_slice = []
    if qop == 0:
        new_slice.append((len(original) - 1, original[-1]))
        return new_slice

    assert isinstance(qop, (list, tuple))
    # filter out strings that are not integers
    qop = [i for i in qop if i.isnumeric()]
    for i in qop:
        new_slice.append((i, original[int(i)]))
    return new_slice


__all__ = [
    "log",
    "arg",
    "checkpointing",
    "loop",
    "iteration",
    "commit",
    "output_buffer",
]



from .clock import Clock
from .constants import *
from . import cli

from .api import *
from .repl import query, dataframe, replay
from . import utils
from . import database

try:
    from IPython import get_ipython

    ip = get_ipython()
    if ip is not None:

        def _callback(*args):
            if output_buffer:
                commit()

        ip.events.register("post_run_cell", _callback)
    else:
        cli.parse_args()
except ImportError:
    cli.parse_args()


conn, cursor = database.conn_and_cursor()
database.create_tables(cursor)
conn.commit()
conn.close()



# type: ignore
from abc import abstractmethod
from itertools import zip_longest
from typing import Any, Callable, Iterable, Optional, Protocol, TypeVar

# https://github.com/PyCQA/pylint/issues/3882
# pylint: disable=unsubscriptable-object


Node = TypeVar("Node")
T = TypeVar("T")


def memoize(orig: Callable[["Adapter", Node], T]) -> Callable[["Adapter", Node], T]:
    memo = {}

    def new(self, n: Node) -> T:
        if (key := id(n)) not in memo:
            memo[key] = orig(self, n)
        return memo[key]

    return new


def materialize(
    orig: Callable[["Adapter", Node], Iterable[T]]
) -> Callable[["Adapter", Node], tuple[T]]:
    def new(self, n: Node) -> tuple[T]:
        return tuple(orig(self, n))

    return new


class BaseAdapter(Protocol[Node]):
    # Implement these for your tree implementation!

    @abstractmethod
    def parent(self, n: Node) -> Optional[Node]:
        raise NotImplementedError

    @abstractmethod
    def children(self, n: Node) -> Iterable[Node]:
        raise NotImplementedError

    @abstractmethod
    def label(self, n: Node) -> str:
        raise NotImplementedError

    @abstractmethod
    def value(self, n: Node) -> Any:
        raise NotImplementedError

    # These should just work, but could be optimized!

    @memoize
    def height(self, n: Node) -> int:
        return 1 + max(map(self.height, self.children(n)), default=0)

    @memoize
    def num_descendants(self, n: Node) -> int:
        return sum(1 + self.num_descendants(c) for c in self.children(n))

    def contains(self, n: Node, t: Node) -> bool:
        while parent := self.parent(n):
            if id(parent) == id(t):
                return True
            n = parent
        return False

    def isomorphic(self, n1: Node, n2: Node) -> bool:
        return all(
            prop(n1) == prop(n2)
            for prop in [self.label, self.value, self.height, self.num_descendants]
        ) and all(
            self.isomorphic(c1, c2)
            for c1, c2 in zip_longest(*map(self.children, [n1, n2]))  # type: ignore
        )

    # These are unlikely to benefit from optimization

    def descendants(self, n: Node) -> Iterable[Node]:
        for child in self.children(n):
            yield from self.descendants(child)
            yield child

    @memoize
    @materialize
    def postorder(self, n: Node) -> Iterable[Node]:
        for child in self.children(n):
            yield from self.postorder(child)
        yield n

    # These are just a debugging / assertion aids

    def root(self, n: Node) -> Node:
        while parent := self.parent(n):
            n = parent
        return n

    def dump(self, n: Node, indent=0):
        return "\n".join(
            [
                "\t" * indent + f"{self.label(n)}: {self.value(n)}",
                *(self.dump(c, indent + 1) for c in self.children(n)),
            ]
        )


# Type export
Adapter = BaseAdapter



import os
from pathlib import Path

from .constants import *
from .clock import Clock
from . import utils
from . import cli

import cloudpickle


def serialize_torch(layers, name, obj):
    import torch
    import torch.nn
    import torch.optim

    if isinstance(obj, torch.nn.Module) or isinstance(obj, torch.optim.Optimizer):
        path = get_shelf() / utils.to_filename(layers, name, ".pth")
        torch.save(obj.state_dict(), path)
        return path.name
    else:
        raise


def serialize_numpy(layers, name, obj):
    import numpy as np

    if isinstance(obj, np.ndarray):
        path = get_shelf() / utils.to_filename(layers, name, ".npy")
        np.save(path, obj)
        return path.name
    else:
        raise


def serialize_scikit(layers, name, obj):
    import pickle  # Scikit-Learn
    import sklearn.base

    sklearn_base_classes = (
        sklearn.base.BaseEstimator,
        sklearn.base.ClassifierMixin,
        sklearn.base.RegressorMixin,
        sklearn.base.ClusterMixin,
        sklearn.base.TransformerMixin,
    )
    if isinstance(obj, sklearn_base_classes) or hasattr(obj, "fit"):
        path = get_shelf() / utils.to_filename(layers, name, ".pkl")
        with open(path, "wb") as f:
            pickle.dump(obj, f)
        return path.name
    else:
        raise


def serialize_pandas(layers, name, obj):
    import pandas as pd

    if isinstance(obj, pd.DataFrame):
        path = get_shelf() / utils.to_filename(layers, name, ".parquet")
        obj.to_parquet(path)
        return path.name
    else:
        raise


def serialize(layers, name, obj):
    try:
        return serialize_torch(layers, name, obj)
    except:
        pass

    try:
        return serialize_scikit(layers, name, obj)
    except:
        pass

    try:
        return serialize_numpy(layers, name, obj)
    except:
        pass

    try:
        return serialize_pandas(layers, name, obj)
    except:
        pass

    with open(
        (path := get_shelf() / utils.to_filename(layers, name, ".pkl")), "wb"
    ) as f:
        cloudpickle.dump(obj, f)
    return path.name


def deserialize(layers, name, obj):
    if (path := get_shelf() / utils.to_filename(layers, name, ".pth")).exists():
        import torch

        obj.load_state_dict(torch.load(path))
    elif (path := get_shelf() / utils.to_filename(layers, name, ".npy")).exists():
        import numpy

        obj[:] = numpy.load(path)
    elif (path := get_shelf() / utils.to_filename(layers, name, ".parquet")).exists():
        import pandas as pd

        obj.iloc[:, :] = pd.read_parquet(path)
    elif (path := get_shelf() / utils.to_filename(layers, name, ".pkl")).exists():
        with open(path, "rb") as f:
            loaded_obj = cloudpickle.load(f)
        obj.clear()
        obj.update(loaded_obj)
    else:
        raise


def get_shelf():
    if not cli.in_replay_mode():
        tstamp = Clock.get_datetime()
    else:
        assert cli.flags.old_tstamp is not None
        tstamp = cli.flags.old_tstamp

    OBJSTORE = Path(HOMEDIR) / "obj_store"
    SHELF = OBJSTORE / PROJID / tstamp
    os.makedirs(SHELF, exist_ok=True)
    return SHELF



from heapq import heapify, heappush, heappop
from typing import Callable, Generic, Iterable, TypeVar
from operator import lt, gt


T, I = TypeVar("T"), TypeVar("I")
Key = Callable[[T], I]


def identity(x):
    return x


class PriorityQ(Generic[T, I]):
    def __init__(self, it: Iterable[T], *, key: Key = identity, reverse=False):
        self._Item = self._itemizer(key, reverse)
        self._heap = list(map(self._Item, it))
        heapify(self._heap)

    def push(self, value: T):
        item = self._Item(value)
        heappush(self._heap, item)

    def pop(self) -> T:
        return heappop(self._heap).value  # type: ignore

    def peek(self) -> T:
        return self._heap[0].value  # type: ignore

    def __len__(self) -> int:
        return len(self._heap)

    def _itemizer(self, key, reverse):
        op = gt if reverse else lt

        class Item:
            def __init__(self, value):
                self.key, self.value = key(value), value

            def __lt__(self, other: "Item"):
                return op(self.key, other.key)

        return Item
